class Node():
    def __init__(self, initdata=None):
        self.data = initdata
        self.next = None
        self.prev = None

    def getData(self):
        return self.data

    def getNext(self):
        return self.next

    def getPrev(self):
        return self.prev

    def setData(self, newdata):
        self.data = newdata

    def setNext(self, newnext):
        self.next = newnext

    def setPrev(self, newprev):
        self.prev = newprev


# ======== 1 UnorderedList ========
# 用链表实现UnorderedList的如下方法：isEmpty, add, search, size, remove, append，index，pop，insert, __len__, __getitem__
# 用于列表字符串表示的__str__方法 (注：__str__里不要使用str(), 用repr()代替
# 用于切片的__getitem__方法
# 选做：UnorderedList(iterable) -> new UnorderedList initialized from iterable's items
# 选做：__eq__, __iter__
class UnorderedList():
    ### 定义与Node相同的访问接口
    def setNext(self, node):
        self.head = node

    getNext = lambda self: self.head

    def __init__(self, argv=None):
        # 声明头尾结点引用
        self.head = self.tail = self

        # 记录链表大小变量
        self.length = 0

        # 若可选参数不为空则依次添加至列表尾（不可迭代则报not iterable错）
        if argv:
            # 逐个加入(append)
            for i in argv:
                self.append(i)

    def _insert_node(self, prv, val):
        '''向两节点间插入新节点'''
        self.length += 1
        nxt = prv.getNext()
        node = Node(val)
        prv.setNext(node)
        node.setNext(nxt)
        if nxt is self:
            self.tail = node
        return node

    def _remove_node(self, prv):
        '''移除指定节点后的节点'''
        node = prv.getNext()
        prv.setNext(node.getNext())
        self.length -= 1
        if self.tail == node:
            self.tail = prv
        return node

    @property
    def _iter_nodes(self):
        '''迭代所有链表节点'''
        # 初始化迭代指针
        iterptr = self.getNext()

        # 在到达末尾前重复迭代，返回data并获取下一个节点
        while iterptr is not self:
            yield iterptr
            iterptr = iterptr.getNext()

        # 结束迭代
        raise StopIteration

    def __iter__(self):
        for node in self._iter_nodes:
            yield node.getData()

    def __eq__(self, other):
        # 若类型不同则不相等
        if type(other) != type(self):
            return False

        # 若长度不同则不相等
        if self.length != len(other):
            return False

        # 逐个比较元素相等
        return all(i == j for i, j in zip(self, other))

    def isEmpty(self):
        # 链表长度为0
        return self.length == 0

    def add(self, item):
        # 插入节点
        new_node = self._insert_node(self, item)

        # 更新链表尾
        if new_node.getNext() is self:
            self.tail = new_node

    def search(self, item):
        # 遍历搜索
        return any(c == item for c in self)

    def size(self):
        # 返回链表大小变量
        return self.length

    def remove(self, item):
        # 由头结点开始搜索
        prv = self
        for nxt in self._iter_nodes:
            if nxt.getData() == item:
                return self._remove_node(prv)
            prv = nxt

        # 搜索结束，无该节点则报错
        raise ValueError('remove(item): %s not in list' % repr(item))

    def append(self, item):
        # 添加至尾端
        self._insert_node(self.tail, item)

    def index(self, value, start=None, stop=None):
        # 在参数缺省时设置默认值
        if not start:
            start = 0
        if not stop:
            stop = self.length

        # 负数索引范围恢复
        if start < 0:
            start += self.length
        if stop < 0:
            stop += self.length

        # 将序号与内容配对并遍历
        for ind, content in zip(range(self.length), self):
            # 序号小于start则跳过
            if ind < start:
                continue

            # 序号不小于stop则结束搜索
            if ind >= stop:
                break

            # 匹配则输出下标
            if content == value:
                return ind

        # 无匹配结果则报错
        raise ValueError('%s is not in list' % repr(value))

    def pop(self, pos=-1):
        # 若列表为空则报错
        if self.length == 0:
            raise IndexError('pop from empty list')

        # 确定pop位置（正数）未超过限制
        if pos < 0:
            pos += self.length
        if pos < 0 or pos >= self.length:
            raise IndexError('pop index out of range')

        # 确定pop位置节点
        prv = self
        for ind in range(pos):
            prv = prv.getNext()

        # 删除节点引用并减小列表长度
        node = self._remove_node(prv)

        # 返回该节点内容
        return node.getData()

    def insert(self, pos, item):
        # 负数pos合法化（与list.insert特性统一）
        if pos < 0:
            pos += self.length

        # 索引超过限制则退化为add或append
        if pos < 0:
            return self.add(item)
        elif pos >= self.length:
            return self.append(item)

        # 确定insert位置节点
        prv = self
        for ind in range(pos):
            prv = prv.getNext()

        # 插入新节点
        self._insert_node(prv, item)

    def __str__(self):
        # 迭代列表并执行repr
        res = map(repr, self)

        # 连接结果并输出
        # 使用type(self).__name__以在继承类中获得准确类名
        return '%s([%s])' % (type(self).__name__, ', '.join(res))

    __repr__ = __str__
    __len__ = size

    def __getitem__(self, argv):
        # 整数index
        if isinstance(argv, int):
            # 确定索引位置（正数）未超过限制
            if argv < 0:
                argv += self.length
            if argv < 0 or argv >= self.length:
                raise IndexError('list index out of range')

            # 确定该位置节点
            ptr = self.getNext()
            for i in range(argv):
                ptr = ptr.getNext()

            # 返回该位置内容
            return ptr.getData()

        # 切片index
        elif isinstance(argv, slice):
            # 列表用于记录切片内容
            # 通过__class__获取实例化函数以在子类继承中保持类型一致
            res = self.__class__()

            # 读取索引起始、终止、步长
            start, stop, step = argv.start, argv.stop, argv.step

            # 合法化step
            if step == None:
                step = 1
            elif step == 0:
                raise ValueError('slice step cannot be zero')

            # 合法化start
            if start == None:
                start = 0 if step > 0 else self.length - 1
            elif start < 0:
                start += self.length

            # 合法化stop
            if stop == None:
                stop = self.length if step > 0 else -1
            elif stop < 0:
                stop += self.length

            # 将内容与下标配对遍历链表
            pairs = zip(range(self.length), self)

            # 根据step正负性决定添加于表头或表尾
            if step > 0:
                #正step添加于表尾
                for ind, content in pairs:
                    if start <= ind < stop and (ind - start) % step == 0:
                        res.append(content)
            else:
                #负step添加于表头
                for ind, content in pairs:
                    if stop < ind <= start and (ind - start) % step == 0:
                        res.add(content)

            # 返回结果
            return res

        # 否则报错
        else:
            raise TypeError('list indices must be integers or slices, not %s' %
                            type(argv).__name__)


# ======== 2 OrderedList ========
# 将OrderedList作为UnorderedList的子类来实现
# 实现ADT OrderedList的所有接口:
# add, remove, search, isEmpty, size, index, pop
# 注：不继承insert, append方法
class OrderedList(UnorderedList):

    # add重写为有序插入，append与add等同
    def add(self, item):
        # 确定插入位置节点
        prv, nxt = self, self.getNext()
        for ind in range(self.length):
            # 若nxt值不小于item则停止搜索
            if nxt.getData() >= item:
                break

            # 否则继续查找
            prv, nxt = nxt, nxt.getNext()

        # 插入新节点
        self._insert_node(prv, item)

    append = add

    # 将insert抛出异常函数屏蔽
    def insert(self, *args, **kwargs):
        raise AttributeError('method disabled')

    # remove, isEmpty, size, pop保序，直接继承

    # search与index按照有序特性优化
    def search(self, item):
        # 遍历列表
        for content in self:
            # 若得到结果则返回真
            if content == item:
                return True

            # 如果结果更大代表全列表无结果
            elif content > item:
                break

        # 搜索完毕无结果则返回假
        return False

    def index(self, value, start=None, stop=None):
        # 在参数缺省时设置默认值
        if not start:
            start = 0
        if not stop:
            stop = self.length

        # 负数索引范围恢复
        if start < 0:
            start += self.length
        if stop < 0:
            stop += self.length

        # 将序号与内容配对并遍历
        for ind, content in zip(range(self.length), self):
            # 序号小于start则跳过
            if ind < start:
                continue

            # 序号不小于stop或值更大则结束搜索
            if ind >= stop or content > value:
                break

            # 匹配则输出下标
            if content == value:
                return ind

        # 无匹配结果则报错
        raise ValueError('%s is not in list' % repr(value))


# ======== 3 ADT Stack & Queue ========
# 用链表实现ADT Stack与ADT Queue的所有接口
class Stack():
    ### 定义与Node相同的访问接口
    def setNext(self, node):
        self.head = node

    getNext = lambda self: self.head

    def __init__(self):
        # 头结点
        self.head = None

        # 记录栈深
        self.length = 0

    def isEmpty(self):
        return self.length == 0

    def push(self, item):
        # 创建新节点
        new_node = Node(item)

        # 加入链表
        new_node.setNext(self.getNext())
        self.setNext(new_node)

        # 更新栈深
        self.length += 1

    def pop(self):
        # 若空栈则报错
        if self.length == 0:
            raise IndexError('pop from empty stack')

        # 获取栈顶内容
        node_to_pop = self.getNext()
        res = node_to_pop.getData()

        # 移除栈顶并更新栈深
        self.setNext(node_to_pop.getNext())
        self.length -= 1

        # 返回内容
        return res

    def peek(self):
        # 获取链表头内容
        return self.getNext().getData()

    def size(self):
        return self.length


class Queue():
    ### 定义与Node相同的访问接口
    def setNext(self, node):
        self.head = node

    getNext = lambda self: self.head

    def __init__(self):
        # 头结点（队列出口）
        self.head = None

        # 队列入口（初始在自身位置）
        self.entry = self

        # 记录队长
        self.length = 0

    def isEmpty(self):
        return self.length == 0

    def enqueue(self, item):
        # 创建新元素
        new_node = Node(item)

        # 入队并更新队尾
        self.entry.setNext(new_node)
        self.entry = new_node

        # 更新长度
        self.length += 1

    def dequeue(self):
        # 若队列为空则报错
        if self.length == 0:
            raise IndexError('queue is empty')

        # 获取队首元素
        qhead = self.getNext()
        res = qhead.getData()

        # 出队并更新长度
        self.setNext(qhead.getNext())
        self.length -= 1
        if self.entry == qhead:
            self.entry = self

        # 返回内容
        return res

    def size(self):
        return self.length


# ======== 4 DoublyLinkedList ========
# 实现双向链表版本的UnorderedList，接口同ADT UnorderedList
# 在节点Node中增加prev变量，引用前一个节点
# 在UnorderedList中增加tail变量，引用列表中最后一个节点
# 增加getTail方法
class DoublyLinkedList(UnorderedList):
    '''
    继承自单链表版无序表
    使用环状结构，便于快速访问头尾节点
    继承方法：
    setNext, getNext,
    __eq__, __len__, __str__, __repr__, __getitem__, 
    index, isEmpty, search, size
    '''

    ### 增加向前访问接口
    def setPrev(self, node):
        self.tail = node

    getPrev = lambda self: self.tail
    getTail = getPrev

    def _insert_node(self, prv, val):
        '''添加反向引用'''
        new_node = super()._insert_node(prv, val)
        new_node.getNext().setPrev(new_node)
        new_node.setPrev(prv)
        return new_node

    def _remove_node(self, prv):
        '''移除指定节点后的节点'''
        node = super()._remove_node(prv)
        node.getNext().setPrev(prv)
        return node
